import pandas as pd
import numpy as np
import datetime as dt

DECIMALS=0
wb_weighting_dB = [
    -8.4,
   -8.3,
   -8.3,
   -8.1,
   -7.6,
   -6.1,
   -3.5,
   -1.1,
   0.2,
   0.5,
   0.2,
   -0.2,
   -0.9,
   -1.8,
   -3.0,
   -4.5,
   -6.2,
   -8.1,
   -10.1,
   -12.4,
   -15.2,
   -18.8,
   -23.2,
   -28.4,
   -34.0,
   -39.8,
   -45.8
]

wb_weighting_factors = [
    0.381,
    0.385,
    0.386,
    0.392,
    0.417,
    0.496,
    0.665,
    0.885,
    1.026,
    1.054,
    1.026,
    0.974,
    0.904,
    0.814,
    0.709,
    0.597,
    0.491,
    0.395,
    0.312,
    0.239,
    0.173,
    0.115,
    0.069,
    0.039,
    0.02,
    0.02,
    0.005
]

class Log:
    def __init__(self, path=""):
        #TODO C:\Users\tonyr\PycharmProjects\pycoustic\.venv2\Lib\site-packages\pycoustic\log.py:15: UserWarning:
        #Parsing dates in %Y/%m/%d %H:%M format when dayfirst=True was specified. Pass `dayfirst=False` or specify a format to silence this warning.
        """
        The Log class is used to store the measured noise data from one data logger.
        The data must be entered in a .csv file with headings in the specific format "Leq A", "L90 125" etc.
        :param path: the file path for the .csv noise data
        """
        self._filepath = path
        self._master = pd.read_csv(
            path,
            index_col="Time",
            parse_dates=["Time"],
            # date_format="%d/%m/%Y %H:%M",  # Explicit format to avoid the dayfirst warning
            dayfirst=True,  # Optional: include for clarity; default is False
        )
        self._master.index = pd.to_datetime(self._master.index)
        self._master = self._master.sort_index(axis=1)
        self._start = self._master.index.min()
        self._end = self._master.index.max()
        self._assign_header()

        # Assign day, evening, night periods
        self._night_start = None
        self._day_start = None
        self._evening_start = None
        self._init_periods()

        # Prepare night-time indices and antilogs
        self._antilogs = self._prep_antilogs()  # Use the antilogs dataframe as input to Leq calculations
        self._master = self._append_night_idx(data=self._master)
        self._antilogs = self._append_night_idx(data=self._antilogs)

        self._decimals = 1

    def _assign_header(self):
        csv_headers = self._master.columns.to_list()
        superheaders = [item.split(" ")[0] for item in csv_headers]
        subheaders = [item.split(" ")[1] for item in csv_headers]
        # Convert numerical subheaders to ints
        for i in range(len(subheaders)):
            try:
                subheaders[i] = float(subheaders[i])
            except Exception:
                continue
        self._master.columns = [superheaders, subheaders]
        self._master.sort_index(axis=1, level=1, inplace=True)

    def _init_periods(self):
        times = {"day": (7, 0), "evening": (23, 0), "night": (23, 0)}
        self._day_start = dt.time(times["day"][0], times["day"][1])
        self._evening_start = dt.time(times["evening"][0], times["evening"][1])
        self._night_start = dt.time(times["night"][0], times["night"][1])


    def _prep_antilogs(self):
        """
        Private method creates a copy dataframe of master, but with dB sound pressure levels presented as antilogs.
        This antilogs dataframe should be used if you want to undertake calculations of Leqs and similar.
        :return:
        """
        return self._master.copy().apply(lambda x: np.power(10, (x / 10)))

    def _append_night_idx(self, data=None):
        """
        Private method appends an additional column of the measurement date and time, but with the early morning
        dates set to the day before.
        e.g.
        the measurement at 16-12-2024 23:57 would stay as is, but
        the measurement at 17-12-2024 00:02 would have a night index of 16-12-2024 00:02
        The logic behind this is that it allows us to process a night-time as one contiguous period, whereas
        Pandas would otherwise treat the two measurements as separate because of their differing dates.
        :param data:
        :return:
        """
        night_indices = data.index.to_list()
        if self._night_start > self._day_start:
            for i in range(len(night_indices)):
                if night_indices[i].time() < self._day_start:
                    night_indices[i] += dt.timedelta(days=-1)
        data["Night idx"] = night_indices
        return data

    def _return_as_night_idx(self, data=None):
        """
        Private method to set the dataframe index as the night_idx. This is used when undertaking data processing for
        night-time periods.
        :param data:
        :return:
        """
        if ("Night idx", "") not in data.columns:
            raise Exception("No night indices in current DataFrame")
        return data.set_index("Night idx")

    def _none_if_zero(self, df):
        if len(df) == 0:
            return None
        else:
            return df

    def _recompute_leq(self, data=None, t="15min", cols=None):
        """
        Private method to recompute shorter Leq measurements as longer ones.
        :param data: Input data (should be in antilog format)
        :param t: The desired Leq period
        :param cols: Which columns of the input data do you wish to recompute?
        :return:
        """
        # Set default mutable args
        if data is None:
            data = self._antilogs
        if cols is None:
            cols = ["Leq", "L90"]
        # Loop through column superheaders and recompute as a longer Leq
        recomputed = pd.DataFrame(columns=data.columns)
        for idx in cols:
            if idx in data.columns:
                recomputed[idx] = data[idx].resample(t).mean().\
                    apply(lambda x: np.round((10 * np.log10(x)), self._decimals))
        return self._none_if_zero(recomputed)

    def _recompute_night_idx(self, data=None, t="15min"):
        """
        Internal method to recompute night index column.
        :param data: input dataframe to be recomputed
        :param t: desired measurement period
        :return: dataframe with night index column recomputed to the desired period
        """
        if data is None:
            raise Exception("No DataFrame provided for night idx")
        if ("Night idx", "") in data.columns:
            data["Night idx"] = data["Night idx"].resample(t).asfreq()
        else:
            data["Night idx"] = self._master["Night idx"].resample(t).asfreq()
            return data

    def _recompute_max(self, data=None, t="15min", pivot_cols=None, hold_spectrum=False):
        """
        Private method to recompute max readings from shorter to longer periods.
        :param data: input data, usually self._master
        :param t: desired measurement period
        :param pivot_cols: how to choose the highest value - this will usually be "Lmax A". This is especially
        important when you want to get specific octave band data for an Lmax event. If you wanted to recompute maxes
        as the events with the highest values at 500 Hz, you could enter [("Lmax", 500)]. Caution: This functionality
        has not been tested
        :param hold_spectrum: if hold_spectrum, the dataframe returned will contain the highest value at each octave
        band over the new measurement period, i.e. like the Lmax Hold setting on a sound level meter.
        If hold_spectrum=false, the dataframe will contain the spectrum for the highest event around the pivot column,
        i.e. the spectrum for that specific LAmax event
        :return: returns a dataframe with the values recomputed to the desired measurement period.
        """
        # Set default mutable args
        if pivot_cols is None:
            pivot_cols = [("Lmax", "A")]
        if data is None:
            data = self._master
        # Loop through column superheaders and recompute over a longer period
        combined = pd.DataFrame(columns=data.columns)
        if hold_spectrum:   # Hold the highest value, in given period per frequency band
            for col in pivot_cols:
                if col in combined.columns:
                    max_hold = data.resample(t)[col[0]].max()
                    combined[col[0]] = max_hold
        else:   # Event spectrum (octave band data corresponding to the highest A-weighted event)
            for col in pivot_cols:
                if col in combined.columns:
                    idx = data[col[0]].groupby(pd.Grouper(freq=t)).max()
                    combined[col[0]] = idx
        return combined

    def _as_multiindex(self, df=None, super=None, name1="Date", name2="Num"):
        subs = df.index.to_list()   # List of subheaders
        # Super will likely be the date
        tuples = [(super, sub) for sub in subs]
        idx = pd.MultiIndex.from_tuples(tuples, names=[name1, name2])
        if isinstance(df, pd.Series):
            df = pd.DataFrame(data=df)
        return df.set_index(idx, inplace=False)
#test
    def get_period(self, data=None, period="days", night_idx=True):
        """
        Private method to get data for daytime, evening or night-time periods.
        :param data: Input data, usually master
        :param period: string, "days", "evenings" or "nights"
        :param night_idx: Bool. Needs to be True if you want to compute contiguous night-time periods. If False,
        it will consider early morning measurements as part of the following day, i.e. the cut-off becomes midnight.
        :return:
        """
        if data is None:
            data = self._master
        if period == "days":
            return data.between_time(self._day_start, self._evening_start, inclusive="left")
        elif period == "evenings":
            return data.between_time(self._evening_start, self._night_start, inclusive="left")
        elif period == "nights":
            if night_idx:
                data = self._return_as_night_idx(data=data)
            return data.between_time(self._night_start, self._day_start, inclusive="left")

    def leq_by_date(self, data, cols=None):
        """
        Private method to undertake Leq calculations organised by date. For contiguous night-time periods crossing
        over midnight (e.g. from 23:00 to 07:00), the input data needs to have a night-time index.
        This method is normally used for calculating Leq over a specific daytime, evening or night-time period, hence
        it is usually passed the output of _get_period()
        :param data: Input data. Must be antilogs, and usually with night-time index
        :param cols: Which columns do you wish to recalculate? If ["Leq"] it will calculate for all subcolumns within
        that heading, i.e. all frequency bands and A-weighted. If you want an individual column, use [("Leq", "A")] for
        example.
        :return: A dataframe of the calculated Leq for the data, organised by dates
        """
        if cols is None:
            cols = ["Leq"]
        return data[cols].groupby(data.index.date).mean().apply(lambda x: np.round((10 * np.log10(x)), self._decimals))

    # ###########################---PUBLIC---######################################
    # ss++
    def get_data(self):
        """
        # Returns a dataframe of the loaded csv
        """
        return self._master
    #ss--

    def get_antilogs(self):
        return self._antilogs


    def as_interval(self, data=None, antilogs=None, t="15min", leq_cols=None, max_pivots=None,
                    hold_spectrum=False):
        """
        Returns a dataframe recomputed as longer periods. This implements the private leq and max recalculations
        :param data: input dataframe, usually master
        :param antilogs: antilog dataframe, used for leq calcs
        :param t: desired output period
        :param leq_cols: which Leq columns to include
        :param max_pivots: which value to pivot the Lmax recalculation on
        :param hold_spectrum: True will be Lmax hold, False will be Lmax event
        :return: a dataframe recalculated to the desired period, with the desired columns
        """
        # Set defaults for mutable args
        if data is None:
            data = self._master
        if antilogs is None:
            antilogs = self._antilogs
        if leq_cols is None:
            leq_cols = ["Leq", "L90"]
        if max_pivots is None:
            max_pivots = [("Lmax", "A")]
        leq = self._recompute_leq(data=antilogs, t=t, cols=leq_cols)
        maxes = self._recompute_max(data=data, t=t, pivot_cols=max_pivots, hold_spectrum=hold_spectrum)
        conc = pd.concat([leq, maxes], axis=1).sort_index(axis=1).dropna(axis=1, how="all")
        conc = self._append_night_idx(data=conc)    # Re-append night indices
        return conc.dropna(axis=0, how="all")

    def get_nth_high_low(self, n=10, data=None, pivot_col=None, all_cols=False, high=True):
        """
        Return a dataframe with the nth-highest or nth-lowest values for the specified parameters.
        This is useful for calculating the 10th-highest or 15th-highest Lmax values, but can be used for other purposes
        :param n: The nth-highest or nth-lowest values to return
        :param data: Input dataframe, usually a night-time dataframe with night-time indices
        :param pivot_col: Tuple of strings,
        Which column to use for the highest-lowest computation. Other columns in the row will follow.
        :param all_cols: Perform this operation over all columns?
        :param high: True for high, False for low
        :return: dataframe with the nth-highest or -lowest values for the specified parameters.
        """
        if data is None:
            data = self._master
        if pivot_col is None:
            pivot_col = ("Lmax", "A")
        nth = None
        if high:
            nth = data.sort_values(by=pivot_col, ascending=False)
        if not high:
            nth = data.sort_values(by=pivot_col, ascending=True)
        nth["Time"] = nth.index.time
        if all_cols:
            return nth.groupby(by=nth.index.date).nth(n-1)
        else:
            return nth[[pivot_col[0], "Time"]].groupby(by=nth.index.date).nth(n-1)

    def get_modal(self, data=None, by_date=True, cols=None, round_decimals=True):
        """
        Return a dataframe with the modal values
        :param data: Input dataframe, usually master
        :param by_date: Bool. Group the modal values by date, as opposed to an overall modal value (currently not
        implemented).
        :param cols: List of tuples of the desired columns. e.g. [("L90", "A"), ("Leq", "A")]
        :param round_decimals: Bool. Round the values to 0 decimal places.
        :return: A dataframe with the modal values for the desired columns, either grouped by date or overall.
        """
        if data is None:
            data = self._master
        if round_decimals:
            data = data.round()
        if cols is None:
            cols = [("L90", "A")]
        if by_date:
            dates = np.unique(data.index.date)
            modes_by_date = pd.DataFrame()
            for date in range(len(dates)):
                date_str = dates[date].strftime("%Y-%m-%d")
                mode_by_date = data[cols].loc[date_str].mode()
                mode_by_date = self._as_multiindex(df=mode_by_date, super=date_str)
                modes_by_date = pd.concat([modes_by_date, mode_by_date])
            return modes_by_date
        else:
            return data[cols].mode()

    def counts(self, data=None, cols=None, round_decimals=True):
        # This does not work with multiple columns
        if data is None:
            data = self._master
        if round_decimals:
            data = data.round(decimals=0)
        if cols is None:
            cols = [("L90", "A")]
        df = data[cols].value_counts()
        df.index = [int(x[0]) for x in df.index]
        return df

    def set_periods(self, times=None):
        """
        Set the daytime, night-time and evening periods. To disable evening periods, simply set it the same
        as night-time.
        :param times: A dictionary with strings as keys and integer tuples as values.
        The first value in the tuple represents the hour of the day that period starts at (24hr clock), and the
        second value represents the minutes past the hour.
        e.g. for daytime from 07:00 to 19:00, evening 19:00 to 23:00 and night-time 23:00 to 07:00,
        times = {"day": (7, 0), "evening": (19, 0), "night": (23, 0)}
        NOTES:
        Night-time must cross over midnight. (TBC experimentally).
        Evening must be between daytime and night-time. To
        :return: None.
        """
        if times is None:
            times = {"day": (7, 0), "evening": (23, 0), "night": (23, 0)}
        self._day_start = dt.time(times["day"][0], times["day"][1])
        self._evening_start = dt.time(times["evening"][0], times["evening"][1])
        self._night_start = dt.time(times["night"][0], times["night"][1])
        # Recompute night indices
        self._master.drop(labels="Night idx", axis=1, inplace=True)
        self._antilogs.drop(labels="Night idx", axis=1, inplace=True)
        self._master = self._append_night_idx(data=self._master)
        self._antilogs = self._append_night_idx(data=self._antilogs)

#C:\Users\tonyr\PycharmProjects\pycoustic\.venv2\Lib\site-packages\pycoustic\log.py:339: PerformanceWarning:
#dropping on a non-lexsorted multi-index without a level parameter may impact performance.

    def get_period_times(self):
        """
        :return: the tuples of period start times.
        """
        return self._day_start, self._evening_start, self._night_start

    def is_evening(self):
        """
        Check if evening periods are enabled.
        :return: True if evening periods are enabled, False otherwise.
        """
        if self._evening_start == self._night_start:
            return False
        else:
            return True

    def get_start(self):
        return self._start

    def get_end(self):
        return self._end


class VibLog(Log):
    """
    This class is a work in progress. Do not use.
    """
    def __init__(self, path, units=None):
        super().__init__(path)
        self._units = units
        if self._units is None:
            self._units = "ms2"
        self._decimals = 12
        self._wb_weighted = self.apply_wb_weighting()

    def head(self):
        return self._master.head()

    def apply_wb_weighting(self, factors=wb_weighting_factors):
        """
        Multiply each column in self._master by the corresponding WB weighting factor and
        store the result in self._wb_weighted.

        Notes:
        - Applies only to numeric columns (non-numeric, e.g. the "Night idx" column, are left unchanged).
        - The mapping is by column order: factor[i] is applied to the i-th numeric column.
        """
        numeric_cols = self._master.select_dtypes(include="number").columns

        if len(numeric_cols) != len(factors):
            raise ValueError(
                f"WB weighting length mismatch: {len(numeric_cols)} numeric columns in _master "
                f"but {len(factors)} factors were provided."
            )

        weighted = self._master.copy()
        weighted.loc[:, numeric_cols] = weighted.loc[:, numeric_cols].mul(factors, axis=1)
        self._wb_weighted = weighted
        return self._wb_weighted

    def sum_across_bands(self, df=None):
        """
        Sum values across band columns and return a DataFrame with:
        - a new column "Sum" containing one summed figure per row
        - the existing "Night idx" column preserved (not included in the sum)
        - the original index preserved
        """
        if df is None:
            df = self._master
        elif df == "wb_weighted":
            df = self._wb_weighted

        night_col = "Night idx"
        if night_col not in df.columns:
            raise ValueError(f"Expected column {night_col!r} in DataFrame.")

        # Sum only numeric columns, but explicitly exclude the Night idx column
        numeric_cols = df.select_dtypes(include="number").columns
        numeric_cols = [c for c in numeric_cols if c != night_col]

        out = df.copy()
        out["Sum"] = out.loc[:, numeric_cols].sum(axis=1)
        return out


    def evdv_from_rms(self, df=None, t=None, col=None):
        """
        Compute eVDV (estimated Vibration Dose Value) from a time history of RMS acceleration.

        This implementation first converts RMS to an estimated peak-equivalent value by:
            a_eq = a_rms * 1.4

        Then computes (discrete):
            eVDV = ( sum_i ( a_eq[i]^4 * t ) )^(1/4)

        Parameters
        ----------
        df : pandas.Series or pandas.DataFrame, optional
            If None, uses self._master.
            If Series, treated as the a_rms time history.
            If DataFrame, `col` must identify the column to use.
        t : float, required
            Assessment interval in seconds.
        col : column label, optional
            Column to use when `df` is a DataFrame (e.g. ("Leq","A") style labels if applicable).

        Returns
        -------
        float
            eVDV in the same acceleration units as the input (e.g. m/s^2).
        """
        if t is None:
            raise ValueError("Parameter 't' (measurement interval in seconds) is required.")
        if t <= 0:
            raise ValueError("Parameter 't' must be a positive number of seconds.")

        data = self._master if df is None else df

        # Extract a 1D series of a_rms values
        if isinstance(data, pd.Series):
            a_rms = data
        else:
            if col is None:
                raise ValueError("When 'df' is a DataFrame, you must provide 'col' to select the a_rms column.")
            a_rms = data[col]

        # Ensure numeric and drop missing
        a_rms = pd.to_numeric(a_rms, errors="coerce").dropna()
        if len(a_rms) == 0:
            raise ValueError("No valid numeric RMS acceleration values available to compute eVDV.")

        a_eq = a_rms * 1.4
        evdv_val = float(((a_eq.pow(4).sum() * float(t)) ** 0.25))
        return evdv_val
    #
    # def get_day_evdv(self):
    #     return self.evdv_from_rms(df=self.get_period(self.sum_across_bands(self.), period="days"), t=1, col="Sum")


import os
import tempfile
from typing import List, Dict

import pandas as pd
import plotly.graph_objects as go
import streamlit as st

from log import *
from survey import *

# Streamlit app config
st.set_page_config(page_title="Pycoustic Acoustic Survey Explorer", layout="wide")

# Graph colour palette config
COLOURS = {
    "Leq A": "#9e9e9e",   # light grey
    "L90 A": "#4d4d4d",   # dark grey
    "Lmax A": "#fc2c2c",  # red
}
# Graph template config
TEMPLATE = "plotly"

if "apply_agg" not in st.session_state:
    st.session_state["apply_agg"] = False
if "period_last" not in st.session_state:
    st.session_state["period_last"] = ""

with st.sidebar:
    # File Upload in expander container
    with st.expander("File Upload", expanded=True):
        files = st.file_uploader(
            "Select one or more CSV files",
            type="csv",
            accept_multiple_files=True,
        )
        if not files:
            st.stop()
    # Integration period entry in expander container
    with st.expander("Integration Period", expanded=True):
        int_period = st.number_input(
            "Insert new integration period (must be larger than data)",
            step=1,
            value=15,
        )
        period_select = st.selectbox(
            "Please select time period",
            ("second(s)", "minute(s)", "hour(s)"),
            index=1,
        )

        # Build the period string
        suffix_map = {"second(s)": "s", "minute(s)": "min", "hour(s)": "h"}
        period = f"{int_period}{suffix_map.get(period_select, '')}"

        # If the period changed since last time, reset the "apply_agg" flag
        if st.session_state["period_last"] != period:
            st.session_state["apply_agg"] = False
            st.session_state["period_last"] = period

        # Button to trigger aggregation for ALL positions
        apply_agg_btn = st.button("Apply Integration Period")
        if apply_agg_btn:
            st.session_state["apply_agg"] = True

# Main Window / Data Load
with st.spinner("Processing Data...", show_time=True):
    # Load each uploaded CSV into a pycoustic Log
    logs: Dict[str, Log] = {}
    for upload_file in files:
        with tempfile.NamedTemporaryFile(delete=False, suffix=".csv") as tmp:
            tmp.write(upload_file.getbuffer())
            path = tmp.name
        try:
            logs[upload_file.name] = Log(path)
        except Exception as err:
            st.error(f"Failed to load `{upload_file.name}` into Pycoustic: {err}")
        finally:
            os.unlink(path)

    # Build Survey and pull summary + spectra
    summary_df = leq_spec_df = lmax_spec_df = None
    summary_error = ""
    if logs:
        try:
            survey = Survey()
            if callable(getattr(survey, "add_log", None)):
                for name, lg in logs.items():
                    survey.add_log(lg, name=name)
            elif hasattr(survey, "_logs"):
                survey._logs = logs

            summary_df = survey.broadband_summary()
            leq_spec_df = getattr(survey, "typical_leq_spectra", lambda: None)()
            lmax_spec_df = getattr(survey, "lmax_spectra", lambda: None)()
        except Exception as err:
            summary_error = str(err)
    else:
        summary_error = "No valid logs loaded."

    # Helper list of “position” names (i.e. filenames)
    pos_list = list(logs.keys())

    # Helper: turn a “spectra” DataFrame into a long‐format table for plotting
    def spectra_to_rows(df: pd.DataFrame, pos_names: List[str]) -> pd.DataFrame | None:
        if df is None:
            return None
        if not isinstance(df.columns, pd.MultiIndex):
            tidy = df.reset_index().rename(columns={df.index.name or "index": "Period"})
            if "Position" not in tidy.columns:
                tidy.insert(0, "Position", pos_names[0] if pos_names else "Pos1")
            return tidy

        # If there is a MultiIndex
        bands = [band for _, band in df.columns][: len({band for _, band in df.columns})]
        set_len = len(bands)
        blocks = []
        for i, pos in enumerate(pos_names):
            start, end = i * set_len, (i + 1) * set_len
            if end > df.shape[1]:
                break
            sub = df.iloc[:, start:end].copy()
            sub.columns = [str(b) for b in bands]
            sub = sub.reset_index().rename(columns={df.index.names[-1] or "index": "Period"})
            if "Position" not in sub.columns:
                sub.insert(0, "Position", pos)
            blocks.append(sub)
        return pd.concat(blocks, ignore_index=True)

    #Create tabs
    ui_tabs = st.tabs(["Summary"] + pos_list)

    #Summary tab
    with ui_tabs[0]:
        st.subheader("Broadband Summary")
        if summary_df is not None:
            st.dataframe(summary_df)
        else:
            st.warning(f"Summary unavailable: {summary_error}")

        # Plot “Typical Leq Spectra” and “Lmax Spectra”, if available
        for title, df_data in (
            ("Typical Leq Spectra", leq_spec_df),
            ("Lmax Spectra", lmax_spec_df),
        ):
            tidy = spectra_to_rows(df_data, pos_list)
            if tidy is None:
                continue

            freq_cols = [c for c in tidy.columns if c not in ("Position", "Period", "A")]
            if freq_cols:
                fig = go.Figure()
                for pos in pos_list:
                    subset = tidy[tidy["Position"] == pos]
                    for _, row in subset.iterrows():
                        period_label = row["Period"]
                        # Cast to string so .lower() is safe
                        period_label_str = str(period_label)
                        mode = (
                            "lines+markers"
                            if period_label_str.lower().startswith("day")
                            else "lines"
                        )
                        label = (
                            f"{pos} {period_label_str}"
                            if len(pos_list) > 1
                            else period_label_str
                        )
                        fig.add_trace(
                            go.Scatter(
                                x=freq_cols,
                                y=row[freq_cols],
                                mode=mode,
                                name=label,
                            )
                        )
                fig.update_layout(
                    template=TEMPLATE,
                    title=f"{title} - Day & Night",
                    xaxis_title="Octave band (Hz)",
                    yaxis_title="dB",
                )
                st.plotly_chart(fig, use_container_width=True)
            else:
                st.warning(f"No frequency columns found for `{title}`.")

    # Position‐Specific Tabs
    for tab, uf in zip(ui_tabs[1:], files):
        with tab:
            log = logs.get(uf.name)
            if log is None:
                st.error(f"Log for `{uf.name}` not found.")
                continue

            # Decide whether to show raw or aggregated data
            if st.session_state["apply_agg"]:
                # 1) Re-aggregate / resample using the chosen period
                try:
                    df_used = log.as_interval(t=period)
                    df_used = df_used.reset_index().rename(
                        columns={df_used.index.name or "index": "Timestamp"}
                    )
                    subheader = "Integrated Survey Data"
                except Exception as e:
                    st.error(f"Failed to apply integration period for `{uf.name}`: {e}")
                    continue
            else:
                # 2) Show the raw data (from log._master) if available
                try:
                    raw_master = log._master  # original DataFrame, indexed by Timestamp
                    df_used = raw_master.reset_index().rename(columns={"Time": "Timestamp"})
                    subheader = "Raw Survey Data"
                except Exception as e:
                    st.error(f"Failed to load raw data for `{uf.name}`: {e}")
                    continue

            # Prepare a flattened‐column header copy JUST FOR PLOTTING
            df_plot = df_used.copy()
            if isinstance(df_plot.columns, pd.MultiIndex):
                flattened_cols = []
                for lvl0, lvl1 in df_plot.columns:
                    lvl0_str = str(lvl0)
                    lvl1_str = str(lvl1) if lvl1 is not None else ""
                    flattened_cols.append(f"{lvl0_str} {lvl1_str}".strip())
                df_plot.columns = flattened_cols

            #  Time‐history Graph (Leq A, L90 A, Lmax A) using df_plot 
            required_cols = {"Leq A", "L90 A", "Lmax A"}
            if required_cols.issubset(set(df_plot.columns)):
                fig = go.Figure()
                fig.add_trace(
                    go.Scatter(
                        x=df_plot["Timestamp"],
                        y=df_plot["Leq A"],
                        name="Leq A",
                        mode="lines",
                        line=dict(color=COLOURS["Leq A"], width=1),
                    )
                )
                fig.add_trace(
                    go.Scatter(
                        x=df_plot["Timestamp"],
                        y=df_plot["L90 A"],
                        name="L90 A",
                        mode="lines",
                        line=dict(color=COLOURS["L90 A"], width=1),
                    )
                )
                fig.add_trace(
                    go.Scatter(
                        x=df_plot["Timestamp"],
                        y=df_plot["Lmax A"],
                        name="Lmax A",
                        mode="markers",
                        marker=dict(color=COLOURS["Lmax A"], size=3),
                    )
                )
                fig.update_layout(
                    template=TEMPLATE,
                    margin=dict(l=0, r=0, t=0, b=0),
                    xaxis=dict(
                        title="Time & Date (hh:mm & dd/mm/yyyy)",
                        type="date",
                        tickformat="%H:%M<br>%d/%m/%Y",
                        tickangle=0,
                    ),
                    yaxis_title="Measured Sound Pressure Level dB(A)",
                    legend=dict(orientation="h", yanchor="top", y=-0.25, xanchor="left", x=0),
                    height=600,
                )
                st.plotly_chart(fig, use_container_width=True)
            else:
                st.warning(f"Required columns {required_cols} missing in {subheader}.")

            # --- Finally, display the TABLE with MultiIndex intact ---
            st.subheader(subheader)
            st.dataframe(df_used, hide_index=True)



import requests
import pandas as pd
import numpy as np
import datetime as dt
import pgeocode as geo
# from .weather import WeatherHistory


DECIMALS=1

pd.set_option('display.max_columns', None)
pd.set_option('display.max_rows', None)

#survey.leq_spectra() bug
#TODO: C:\Users\tonyr\PycharmProjects\pycoustic\.venv1\Lib\site-packages\pycoustic\survey.py:287: FutureWarning: The behavior of pd.concat with len(keys) != len(objs) is deprecated. In a future version this will raise instead of truncating to the smaller of the two sequences combi = pd.concat(all_pos, axis=1, keys=["UA1", "UA2"])
#TODO: Survey should make a deep copy of Log objects. Otherwise setting time periods messes it up for other instances.

class Survey:
    """
    Survey Class is an overarching class which takes multiple Log objects and processes and summarises them together.
    This should be the main interface for user interaction with their survey data.
    """

    # ###########################---PRIVATE---######################################

    def __init__(self):
        self._logs = {}
        self._weather = WeatherHistory()

    def _insert_multiindex(self, df=None, super=None, name1="Position", name2="Date"):
        subs = df.index.to_list()   # List of subheaders (dates)
        # Super should be the position name (key from master dictionary)
        tuples = [(super, sub) for sub in subs]
        idx = pd.MultiIndex.from_tuples(tuples, names=[name1, name2])
        return df.set_index(idx, inplace=False)

    def _insert_header(self, df=None, new_head_list=None, header_idx=None):
        if df is None or new_head_list is None:
            return df

        ncols = df.shape[1]

        # Allow scalar broadcast for convenience
        if isinstance(new_head_list, (str, int, float)):
            new_head_list = [new_head_list] * ncols

        if len(new_head_list) != ncols:
            raise ValueError(
                f"new_head_list length ({len(new_head_list)}) must equal number of columns ({ncols})."
            )

        cols = list(df.columns)
        # Normalize columns into list of tuples
        if isinstance(df.columns, pd.MultiIndex):
            tuples = cols  # already list of tuples
        else:
            tuples = [(c,) for c in cols]

        # Build arrays per level and insert new header list
        arrays = [list(x) for x in zip(*tuples)]
        arrays.insert(header_idx, list(new_head_list))
        df.columns = pd.MultiIndex.from_arrays(arrays)
        return df

    # ###########################---PUBLIC---######################################

    def set_periods(self, times=None):
        """
        Set the daytime, evening and night-time periods of all Log objects in the Survey.
        To disable evening periods, simply set it the same as night-time.
        :param times: A dictionary with strings as keys and integer tuples as values.
        The first value in the tuple represents the hour of the day that period starts at (24hr clock), and the
        second value represents the minutes past the hour.
        e.g. for daytime from 07:00 to 19:00, evening 19:00 to 23:00 and night-time 23:00 to 07:00,
        times = {"day": (7, 0), "evening": (19, 0), "night": (23, 0)}
        NOTES:
        Night-time must cross over midnight. (TBC experimentally).
        Evening must be between daytime and night-time. To
        :return: None.
        """
        if times is None:
            times = {"day": (7, 0), "evening": (23, 0), "night": (23, 0)}
        for key in self._logs.keys():
            self._logs[key].set_periods(times=times)

    def add_log(self, data=None, name=""):
        """
        Add a Log object to the Survey object.
        :param data: Initialised Log object
        :param name: Name of the position, e.g. "A1"
        :return: None.
        """
        self._logs[name] = data

    def get_periods(self):
        """
        Check the currently-set daytime, evening and night-time periods for each Log object in the Survey.
        :return: Tuples of start times.
        """
        periods = {}
        for key in self._logs.keys():
            periods[key] = self._logs[key].get_period_times()
        return periods

    def broadband_summary(self, leq_cols=None, max_cols=None, lmax_n=10, lmax_t="2min"):
        """
        Get a dataframe summarising the parameters relevant to assessment of internal ambient noise levels in
        UK residential property assessments. Daytime and night-time Leqs, and nth-highest Lmax values all presented
        in a succinct table. These will be summarised as per the daytime, evening and night-time periods set (default
        daytime 07:00 to 23:00 and night-time 23:00 to 07:00).
        The date of the Lmax values are presented for the night-time period beginning on that date. i.e. an Lmax
        on 20/12/2024 would have occurred in the night-time period starting on that date and ending the following
        morning.
        :param leq_cols: List of tuples. The columns on which to perform Leq calculations. This can include L90
        columns, or spectral values. e.g.  leq_cols = [("Leq", "A"), ("L90", "125")]
        :param max_cols: List of tuples. The columns on which to get the nth-highest values.
        Default max_cols = [("Lmax", "A")]
        :param lmax_n: Int. The nth-highest value for max_cols. Default 10 for 10th-highest.
        :param lmax_t: String. This is the time period over which to compute nth-highest Lmax values.
        e.g. "2min" computes the nth-highest Lmaxes over 2-minute periods. Note that the chosen period must be
        equal to or more than the measurement period. So you cannot measure in 5-minute periods and request 2-minute
        Lmaxes.
        :return: A dataframe presenting a summary of the Leq and Lmax values requested.
        """
        combi = pd.DataFrame()
        if leq_cols is None:
            leq_cols = [("Leq", "A")]
        if max_cols is None:
            max_cols = [("Lmax", "A")]

        for key, lg in self._logs.items():
            period_blocks = []
            period_names = []

            # Day
            days = lg.leq_by_date(lg.get_period(data=lg.get_antilogs(), period="days"), cols=leq_cols)
            days.sort_index(inplace=True)
            period_blocks.append(days)
            period_names.append("Daytime")

            # Evening (optional)
            if lg.is_evening():
                evenings = lg.leq_by_date(lg.get_period(data=lg.get_antilogs(), period="evenings"), cols=leq_cols)
                evenings.sort_index(inplace=True)
                period_blocks.append(evenings)
                period_names.append("Evening")

            # Night Leq
            nights = lg.leq_by_date(lg.get_period(data=lg.get_antilogs(), period="nights"), cols=leq_cols)
            nights.sort_index(inplace=True)
            period_blocks.append(nights)
            period_names.append("Night-time")

            # Night max
            maxes = lg.as_interval(t=lmax_t)
            maxes = lg.get_period(data=maxes, period="nights", night_idx=True)
            maxes = lg.get_nth_high_low(n=lmax_n, data=maxes)[max_cols]
            maxes.sort_index(inplace=True)
            try:
                maxes.index = pd.to_datetime(maxes.index)
                maxes.index = maxes.index.date
            except Exception as e:
                print(f"Error converting index to date: {e}")
            maxes.index.name = None
            period_blocks.append(maxes)
            period_names.append("Night-time")

            # Build Period as a proper column level to avoid manual header length mismatches
            summary = pd.concat(objs=period_blocks, axis=1, keys=period_names, names=["Period"])

            # Add the per-log super level (kept as before)
            summary = self._insert_multiindex(df=summary, super=key)

            # Stack logs by rows; columns remain aligned across logs
            combi = pd.concat(objs=[combi, summary], axis=0)

        # No manual header insertion needed anymore; Period is already a column level
        return combi.round(decimals=DECIMALS)
#test
    def modal(self, cols=None, by_date=False, day_t="60min", evening_t="60min", night_t="15min"):
        """
        Get a dataframe summarising Modal L90 values for each time period, as suggested by BS 4142:2014.
        Currently, this method rounds the values to 0 decimal places by default and there is no alternative
        implementation.
        Note that this function will estimate L90s as a longer value by performing an Leq computation on them.
        The measured data in Logs must be smaller than or equal to the desired period, i.e. you can't measure in 15-
        minute periods and request 5-minute modal values.
        :param cols: List of tuples of the columns desired. This does not have to be L90s, but can be any column.
        :param by_date: Bool. If True, group the modal values by date. If False, present one modal value for each
        period.
        :param day_t:  String. Measurement period T. i.e. daytime measurements will compute modal values of
        L90,60min by default.
        :param evening_t: String. Measurement period T. i.e. evening measurements will compute modal values of
        L90,60min by default, unless evenings are disabled (which they are by default).
        :param night_t: Measurement period T. i.e. night-time measurements will compute modal values of
        L90,15min by default.
        :return: A dataframe of modal values for each time period.
        """
        if cols is None:
            cols = [("L90", "A")]
        combi = pd.DataFrame()
        period_headers = []
        for key in self._logs.keys():
            # Key is the name of the measurement position
            log = self._logs[key]
            pos_summary = []
            # Daytime
            period_headers = ["Daytime"]
            days = log.get_modal(data=log.get_period(data=log.as_interval(t=day_t), period="days"), by_date=by_date, cols=cols)
            days.sort_index(inplace=True)
            pos_summary.append(days)
            # Evening
            if log.is_evening():
                period_headers.append("Evening")
                evenings = log.get_modal(data=log.get_period(data=log.as_interval(t=evening_t), period="evenings"), by_date=by_date, cols=cols)
                evenings.sort_index(inplace=True)
                pos_summary.append(evenings)
            # Night time
            nights = log.get_modal(data=log.get_period(data=log.as_interval(t=night_t), period="nights"), by_date=by_date, cols=cols)
            nights.sort_index(inplace=True)
            pos_summary.append(nights)
            period_headers.append("Night-time")
            pos_df = pd.concat(pos_summary, axis=1)
            pos_df = self._insert_multiindex(pos_df, super=key)
            combi = pd.concat([combi, pos_df], axis=0)
        combi = self._insert_header(df=combi, new_head_list=period_headers, header_idx=0)
        combi.rename_axis(index={'Date': '#'}, inplace=True)
        return combi

    def counts(self, cols=None, day_t="60min", evening_t="60min", night_t="15min"):
        """
        Returns counts for each time period. For example, this can return the number of L90 occurrences at each decibel
        level for daytime and night-time periods.
        :param cols: Which columns to consider. Default (L90, A).
        :param day_t: Daytime averaging period. Default 60min.
        :param evening_t: Evening average period. Default 60min.
        :param night_t: Night-time average period. Default 15min.
        :return: Returns a dataframe of counts for each time period.
        """
        if cols is None:
            cols = [("L90", "A")]
        combi = pd.DataFrame()
        period_headers = []
        for key in self._logs.keys():
            # Key is the name of the measurement position
            log = self._logs[key]
            pos_summary = []
            # Daytime
            # period_headers = ["Daytime"]
            days = log.counts(data=log.get_period(data=log.as_interval(t=day_t), period="days"), cols=cols)
            days.sort_index(inplace=True)
            days.name = "Daytime"
            pos_summary.append(days)
            # Evening
            if log.is_evening():
                # period_headers.append("Evening")
                evenings = log.counts(data=log.get_period(data=log.as_interval(t=evening_t), period="evenings"), cols=cols)
                evenings.sort_index(inplace=True)
                evenings.name = "Evening"
                pos_summary.append(evenings)
            # Night time
            nights = log.counts(data=log.get_period(data=log.as_interval(t=night_t), period="nights"), cols=cols)
            nights.sort_index(inplace=True)
            pos_summary.append(nights)
            nights.name = "Night-time"
            # period_headers.append("Night-time")
            pos_df = pd.concat(pos_summary, axis=1)
            # ensure integer dtype for every period block
            pos_df = pos_df.fillna(0).astype("int64")
            pos_df = self._insert_multiindex(pos_df, super=key)
            combi = pd.concat([combi, pos_df], axis=0)
        # combi = self._insert_header(df=combi, new_head_list=period_headers, header_idx=0)
        combi.sort_index(inplace=True)
        combi.rename_axis(index={'Date': 'dB'}, inplace=True)
        return combi.fillna(0).astype("int64")


    def lmax_spectra(self, n=10, t="2min", period="nights"):
        """
        Get spectral data for the nth-highest Lmax values during a given time period.
        This computes Lmax Event spectra. Lmax Hold spectra has not yet been implemented.
        Assumptions and inputs as per Survey.resi_summary() method.
        IMPORTANT: The dates of the Lmax values are presented for the night-time period beginning on that date.
        This means that for early morning timings, the date is behind by one day.
        e.g. an Lmax presented as occurring at 20/12/2024 at 01:22 would have occurred at 21/12/2024 at 01:22.
        :param n: Int. Nth-highest Lmax. Default 10th-highest.
        :param t: String. This is the time period over which to compute nth-highest Lmax values.
        e.g. "2min" computes the nth-highest Lmaxes over 2-minute periods. Note that the chosen period must be
        equal to or more than the measurement period. So you cannot measure in 5-minute periods and request 2-minute
        Lmaxes.
        :param period: String. "days", "evenings" or "nights"
        :return: Dataframe of nth-highest Lmax Event spectra.
        """
        combi = pd.DataFrame()
        # TODO: The night-time timestamp on this is sometimes out by a minute.
        for key in self._logs.keys():
            log = self._logs[key]
            combined_list = []
            maxes = log.get_nth_high_low(n=n, data=log.get_period(data=log.as_interval(t=t), period=period))[["Lmax", "Time"]]
            maxes.sort_index(inplace=True)
            combined_list.append(maxes)
            summary = pd.concat(objs=combined_list, axis=1)
            summary = self._insert_multiindex(df=summary, super=key)
            combi = pd.concat(objs=[combi, summary], axis=0)
        return combi.round(decimals=DECIMALS)

    def leq_spectra(self, leq_cols=None):
        """
        Compute Leqs over daytime, evening and night-time periods.
        This is an overall Leq, and does not group Leqs by date.
        :param leq_cols: List of strings or List of Tuples.
        For all Leq columns, use ["Leq"]. For specific columns, use list of tuples [("Leq", "A"), ("Leq", 125)]
        :return: A dataframe with a continuous Leq computation across dates, for each time period.
        """
        all_pos = []
        labels = []

        if leq_cols is None:
            leq_cols = ["Leq"]

        for label, log in self._logs.items():
            # Day
            days = log.get_period(data=log.get_antilogs(), period="days")
            days = days[leq_cols].apply(lambda x: np.round(10 * np.log10(np.mean(x)), DECIMALS))

            # Night-time
            nights = log.get_period(data=log.get_antilogs(), period="nights")
            nights = nights[leq_cols].apply(lambda x: np.round(10 * np.log10(np.mean(x)), DECIMALS))

            # Evening (if applicable)
            if log.is_evening():
                evenings = log.get_period(data=log.get_antilogs(), period="evenings")
                evenings = evenings[leq_cols].apply(lambda x: np.round(10 * np.log10(np.mean(x)), DECIMALS))
                df = pd.concat([days, evenings, nights], axis=1, keys=["Daytime", "Evening", "Night-time"])
            else:
                df = pd.concat([days, nights], axis=1, keys=["Daytime", "Night-time"])

            all_pos.append(df)
            labels.append(label)

        if not all_pos:
            return pd.DataFrame()

        # Concatenate across logs; keys match number of objects (no FutureWarning)
        combi = pd.concat(all_pos, axis=1, keys=labels)
        return combi.transpose().round(decimals=DECIMALS)

    def get_start_end(self):
        starts = [self._logs[key].get_start() for key in self._logs.keys()]
        ends = [self._logs[key].get_end() for key in self._logs.keys()]
        return min(starts), max(ends)

    def weather_config(self, **kwargs):
        """
        Configure the embedded WeatherHistory using the Survey's overall start/end
        (earliest log start, latest log end). Additional WeatherHistory kwargs may
        be provided by the caller (e.g. interval_hours, api_key, country, postcode, units, recompute).
        """
        if not self._logs:
            raise ValueError("No logs in Survey. Add Log objects before configuring weather.")
        start, end = self.get_start_end()
        return self._weather.configure(start=start, end=end, **kwargs)

    def weather_compute(self, **kwargs):
        """
        Compute weather history via the embedded WeatherHistory for the Survey's overall
        start/end (earliest log start, latest log end). Caller may pass through any
        WeatherHistory.compute kwargs (e.g. drop_cols, recompute, timeout_s).

        If WeatherHistory is not yet configured, this will configure it first using the
        Survey's start/end plus any kwargs intended for configure(...).
        """
        if not self._logs:
            raise ValueError("No logs in Survey. Add Log objects before computing weather.")
        if self._weather is None:
            raise ValueError("WeatherHistory object is missing from this Survey instance.")

        # If not configured yet, configure first. Split kwargs between configure/compute by name.
        if getattr(self._weather, "_config", None) is None:
            cfg_keys = {"interval_hours", "api_key", "country", "postcode", "units", "recompute"}
            cfg_kwargs = {k: v for k, v in kwargs.items() if k in cfg_keys}
            self.weather_config(**cfg_kwargs)

        compute_keys = {"drop_cols", "recompute", "timeout_s"}
        compute_kwargs = {k: v for k, v in kwargs.items() if k in compute_keys}
        return self._weather.compute(**compute_kwargs)

    def get_weather_hist(self):
        """Return cached computed weather dataframe (or None if not computed yet)."""
        return self._weather.get_weather_history()

    def get_weather_raw(self):
        """Return cached raw API responses (or None if not computed yet)."""
        return self._weather.get_raw_responses()


from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union


@dataclass(frozen=True)
class WeatherConfig:
    start: dt.datetime
    end: dt.datetime
    interval_hours: int
    api_key: str
    country: str
    postcode: str
    units: str = "metric"


class WeatherHistory:
    """
    Fetches historic weather snapshots from OpenWeather One Call 3.0 Time Machine endpoint
    at a fixed interval between start and end (inclusive-ish; end is an upper bound).

    Design:
    - Configure once with (start, end, interval, api_key, country, postcode, units)
    - Compute once and cache the resulting dataframe
    - Subsequent compute() calls return cached data unless recompute=True
    """

    ONECALL_TIMEMACHINE_URL = "https://api.openweathermap.org/data/3.0/onecall/timemachine"
    DEFAULT_DROP_COLS = ("sunrise", "sunset", "feels_like", "dew_point", "visibility")

    def __init__(self) -> None:
        self._config: Optional[WeatherConfig] = None
        self._latlon: Optional[Tuple[float, float]] = None
        self._data: Optional[pd.DataFrame] = None
        self._raw_responses: Optional[List[Dict[str, Any]]] = None

    # ----------------------------
    # Configuration / caching
    # ----------------------------
    def configure(
        self,
        *,
        start: Union[str, dt.datetime],
        end: Union[str, dt.datetime],
        interval_hours: int = 12,
        api_key: Optional[str] = None,
        country: str = "GB",
        postcode: str = "WC1",
        units: str = "metric",
        recompute: bool = False,
    ) -> "WeatherHistory":
        """
        Store config. If config changes, cached data is invalidated automatically.
        If recompute=True, cached data is cleared even if config is unchanged.
        """
        if not api_key:
            raise ValueError("api_key is required")

        start_dt = self._coerce_datetime(start)
        end_dt = self._coerce_datetime(end)

        if end_dt <= start_dt:
            raise ValueError("end must be after start")

        if interval_hours <= 0:
            raise ValueError("interval_hours must be a positive integer")

        new_cfg = WeatherConfig(
            start=start_dt,
            end=end_dt,
            interval_hours=int(interval_hours),
            api_key=str(api_key),
            country=str(country),
            postcode=str(postcode),
            units=str(units),
        )

        if recompute or (self._config != new_cfg):
            self._config = new_cfg
            self._latlon = None
            self._data = None
            self._raw_responses = None
        else:
            self._config = new_cfg

        return self

    def compute(
        self,
        *,
        drop_cols: Optional[List[str]] = None,
        recompute: bool = False,
        timeout_s: int = 30,
    ) -> pd.DataFrame:
        """
        Compute and return a dataframe of weather snapshots.
        Uses cached value unless recompute=True.
        """
        if self._config is None:
            raise ValueError("WeatherHistory is not configured. Call configure(...) first.")

        if self._data is not None and not recompute:
            return self._data

        lat, lon = self._get_latlon(country=self._config.country, postcode=self._config.postcode)

        timestamps = self._construct_timestamps(
            start=self._config.start,
            end=self._config.end,
            interval_hours=self._config.interval_hours,
        )

        rows: List[pd.Series] = []
        raw: List[Dict[str, Any]] = []

        for ts in timestamps:
            payload = self._call_openweather(
                lat=lat,
                lon=lon,
                timestamp=ts,
                api_key=self._config.api_key,
                units=self._config.units,
                timeout_s=timeout_s,
            )
            raw.append(payload)

            row = self._parse_timemachine_payload(payload)
            rows.append(pd.Series(row))

        df = pd.concat(rows, axis=1).T if rows else pd.DataFrame()

        # Convert unix seconds to datetimes where present
        for col in ("dt", "sunrise", "sunset"):
            if col in df.columns:
                df[col] = df[col].apply(
                    lambda x: dt.datetime.fromtimestamp(int(x)) if pd.notna(x) else x
                )

        cols_to_drop = list(self.DEFAULT_DROP_COLS) if drop_cols is None else list(drop_cols)
        existing = [c for c in cols_to_drop if c in df.columns]
        if existing:
            df = df.drop(columns=existing)

        self._data = df
        self._raw_responses = raw
        return self._data

    # ----------------------------
    # Accessors
    # ----------------------------
    def get_weather_history(self) -> Optional[pd.DataFrame]:
        """Return cached dataframe (or None if not computed yet)."""
        return self._data

    def get_raw_responses(self) -> Optional[List[Dict[str, Any]]]:
        """Return cached raw API payloads (or None if not computed yet)."""
        return self._raw_responses

    # ----------------------------
    # Internals
    # ----------------------------
    @staticmethod
    def _coerce_datetime(value: Union[str, dt.datetime]) -> dt.datetime:
        if isinstance(value, dt.datetime):
            return value
        if isinstance(value, str):
            # Accept the existing convention used elsewhere in your project
            return dt.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        raise TypeError("start/end must be datetime or 'YYYY-mm-dd HH:MM:SS' string")

    def _get_latlon(self, *, country: str, postcode: str) -> Tuple[float, float]:
        if self._latlon is not None:
            return self._latlon

        nomi = geo.Nominatim(country)
        rec = nomi.query_postal_code(postcode)

        lat = getattr(rec, "latitude", None)
        lon = getattr(rec, "longitude", None)

        if lat is None or lon is None or pd.isna(lat) or pd.isna(lon):
            raise ValueError(f"Could not resolve postcode to lat/lon: {country=} {postcode=}")

        self._latlon = (float(lat), float(lon))
        return self._latlon

    @staticmethod
    def _construct_timestamps(*, start: dt.datetime, end: dt.datetime, interval_hours: int) -> List[int]:
        """
        Build timestamps starting at 'start', stepping by interval_hours, strictly less than end.
        """
        timestamps: List[int] = [int(start.timestamp())]
        next_time = start + dt.timedelta(hours=interval_hours)
        while next_time < end:
            timestamps.append(int(next_time.timestamp()))
            next_time += dt.timedelta(hours=interval_hours)
        return timestamps

    def _call_openweather(
        self,
        *,
        lat: float,
        lon: float,
        timestamp: int,
        api_key: str,
        units: str,
        timeout_s: int,
    ) -> Dict[str, Any]:
        params = {
            "lat": lat,
            "lon": lon,
            "dt": int(timestamp),
            "appid": api_key,
            "units": units,
        }

        resp = requests.get(self.ONECALL_TIMEMACHINE_URL, params=params, timeout=timeout_s)
        # Raise for 4xx/5xx with a helpful message
        try:
            resp.raise_for_status()
        except requests.HTTPError as e:
            msg = f"OpenWeather request failed: status={resp.status_code} body={resp.text[:500]}"
            raise requests.HTTPError(msg) from e

        return resp.json()

    @staticmethod
    def _parse_timemachine_payload(payload: Dict[str, Any]) -> Dict[str, Any]:
        """
        OpenWeather Time Machine returns a payload with a 'data' array.
        We take the first element as the snapshot.

        Keeps 'weather' out by default; flattens some common nested fields.
        """
        data = payload.get("data", [])
        if not data:
            raise ValueError("OpenWeather payload missing 'data' items")

        snap = dict(data[0])  # shallow copy

        # Flatten rain/snow dicts (e.g. {"1h": 0.25}) into rain_1h
        for k in ("rain", "snow"):
            if k in snap and isinstance(snap[k], dict):
                for subk, val in snap[k].items():
                    snap[f"{k}_{subk}"] = val
                del snap[k]

        # Optionally flatten weather[0].main/description/icon into columns
        wx = snap.get("weather")
        if isinstance(wx, list) and wx:
            w0 = wx[0] if isinstance(wx[0], dict) else None
            if w0:
                for key in ("main", "description", "icon"):
                    if key in w0:
                        snap[f"weather_{key}"] = w0[key]
        if "weather" in snap:
            del snap["weather"]

        return snap